#
# parser generater for ecpg
# call with backend parser as stdin
#
# Copyright (c) 2007-2008, PostgreSQL Global Development Group
#
# Written by Mike Aubury <mike.aubury@aubit.com>
#	     Michael Meskes <meskes@postgresql.org>
#
# Placed under the same license as PostgreSQL.
#

function add_to_buffer(buffer, str)  {
	buff[buffer, buffcnt[buffer]++]=str
}

function dump(buffer)  {
	print "/* " buffer " */"
	for (a=0; a<buffcnt[buffer]; a++) {
		print buff[buffer,a]
	}
}

# Dump the action for a rule - 
# mode indicates if we are processing the 'stmt:' rule (mode==0 means normal,  mode==1 means stmt:)
# flds are the fields to use. These may start with a '$' - in which case they are the result of a previous non-terminal
#                             if they dont start with a '$' then they are token name
#
# len is the number of fields in flds...
# leadin is the padding to apply at the beginning (just use for formatting)
function dump_fields(mode, flds, len, leadin) {

	if (mode==0) { #Normal 
		
		if (len==0) { # We have no fields ? 
			ln=leadin "  { $$=EMPTY; }"
			add_to_buffer("rules", ln);
		} else {
			# Go through each field and try to 'aggregate' the tokens into a single 'make_str' where possible
			cnt=0;
			for (z=0;z<len;z++) { 

				if (substr(flds[z],1,1)=="$") {
					flds_new[cnt++]=flds[z];
					continue
				}

				if (z==len-1) {
					flds_new[cnt++]="make_str(\"" flds[z] "\")";
					continue
				}

				str=flds[z]

				while (1) {
					if (substr(flds[z+1],1,1)=="$" || z>=len-1) { # We're at the end...
						flds_new[cnt++]="make_str(\"" str "\")";
						break;
					}
					z++;
					str=str " " flds[z]
				}
			}
			
			# So - how many fields did we end up with ? 
			if (cnt==1) {
				# Straight assignement
				str=leadin "   { $$ = " flds_new[0] "; }" 
				add_to_buffer("rules", str);
			} else {
				# Need to concatenate the results to form
				# our final string
				str="   { $$ = cat_str(" cnt  ","

				for (z=0;z<cnt;z++) { 
						if (z) {
							str=str ","
						}
						str=str flds_new[z]; 	
				}
				str=leadin str ");}"
				add_to_buffer("rules",str)
			}
		}
	}

	if (mode==1) {
		if (len) {
			# Just output the statement....
			ln=leadin "   { output_statement($1, 0, ECPGst_normal); }"
			add_to_buffer("rules",ln)
			
		} else {
			ln=leadin "{ $$ = NULL; }"
			add_to_buffer("rules",ln)
			
		}
	}
}

function dump_line(stmt_mode, fields, field_count, leading_spaces)  {
	line = leading_spaces line
	add_to_buffer("rules", line)
	if (buffcnt["literal"] > 0) {
		for (a=0; a<buffcnt["literal"]; a++) {
			add_to_buffer("rules", buff["literal",a])
		}
		buffcnt["literal"]=0;
	} else {
		dump_fields(stmt_mode, fields, field_count, leading_spaces);
	}
}

BEGIN {
	leading_spaces=""
	copymode=0;
	brace_indent=0;
	yaccmode=0;

	# Some symbols have to be replaced by other symbols
	replace_token["BCONST"]="ecpg_bconst"
	replace_token["FCONST"]="ecpg_fconst"
	replace_token["IDENT"]="ecpg_ident"
	replace_token["PARAM"]="ecpg_param"
	replace_string["WITH_TIME"]="with time"
	replace_string["NULLS_FIRST"]="nulls first"
	replace_string["NULLS_LAST"]="nulls last"
	replace_string["TYPECAST"]="::"

	# Specific types for specific non-terminals - never include the ':'
	# ECPG-only types are defined in ecpg-types
	types["PrepareStmt"]="<prep>"
 	types["opt_array_bounds"]="<index>"
}

/\@ECPGTOKENS/ { 
	tokenmode=1;
	next
}

/\@ECPGINCLUDE/ {
		includestream	=$3
		includefilename	=$4
		includetype	=$5
		includeblock	=$6
		# if this block starts with a new rule, we have to dump the last one first
		if (includetype == "rule") {
			dump_line(stmt_mode, fields, field_count, leading_spaces);
			field_count = 0;
			line = "";
			infield = 1;
			copy = 0;
		}
		else if (includetype == "block") {
			copy = 0;
		}
		else
			copy = 1;
		while (getline < includefilename)  {
			if ((includetype == "block" || includetype == "rule") && $1 == "ECPG:") {
				if ($2 == includeblock) copy = 1;
				else copy = 0;
			}
			else {
				if (copy == 1) add_to_buffer(includestream, $0);
			}
		}
		close(includefilename)
	next
}


/\@ECPGON/ {
	copymode=1;
	brace_indent=0;
	next
}


/\@ECPGOFF/ {
	copymode=0;
	brace_indent=0;
	next
}

/^%%/	{
		tokenmode=0;
		yaccmode++; infield=0; fieldcount=0
	}

{
	if (comment) {
			commentstr=commentstr "\n"
	}

	if (tokenmode==1) {
		s=$0
		# Make sure  any braces appear on separate lines
		gsub("{","\n{\n",s)
		gsub("}","\n}\n",s)

		sfind="[/][*]"
		srepl="/* "
		gsub(sfind,srepl,s)
		sfind="[*][/]"
		srepl=" */"
		gsub(sfind,srepl,s)
		# Now split the line into individual fields
		n=split(s,arr)
		str=""
		for (a=1;a<=n;a++) {
			if (arr[a]=="/*") {comment++; continue}
			if (arr[a]=="*/") {comment--; continue}
			if (comment) {continue;}	

			if (arr[a]=="%token") 		{type=1; }
			if (arr[a]=="%left") 		{type=2; }
			if (arr[a]=="%right") 		{type=3; }
			if (arr[a]=="%nonassoc") 	{type=4; }
			if (substr(arr[a],1,1)=="<")  continue; # its a type

			tokens[arr[a]]=type;
			token_order[token_cnt++]=arr[a];

			if (str=="") {
				str=arr[a]
			} else {
				str=str " " arr[a]
			}

		}
		add_to_buffer("orig_tokens",str)
		next;
	}

	# Dont worry about anything if we're not in copy mode..
	if (copymode==0) {
		next
	}
	if (yaccmode<1) {
		add_to_buffer("rules",$0)
		next
	}

	s=$0
	# Make sure any braces are split
	gsub("{"," { ",s) gsub("}"," } ",s)
	# Any comments are split
		sfind="[/][*]"
		srepl="/* "
		gsub(sfind,srepl,s)
		sfind="[*][/]"
		srepl=" */"
		gsub(sfind,srepl,s)

	# Set the current indent by calculating the leading spaces
	ln_leading_spaces="";
	for (a=1; a<length(s); a++) {
		if (substr(s,a,1)=="\t" || substr(s,a,1)==" ") {
			continue;
		} else {
			ln_leading_spaces=substr(s,1,a-1)
			break;
		}
	}

	# Now split the line into individual fields
	n=split(s, arr);
	prec = 0;
	add = 0;

	# Go through each field in turn
	for (fieldIndexer=1; fieldIndexer<=n; fieldIndexer++) {
		if (arr[fieldIndexer]=="*/" && comment) { 
			commentstr=commentstr " */"
			comment=0;
			add = 0;
			add_to_buffer("rules", commentstr)
			commentstr=""
			continue
		}

		if (comment) {
			if (!add) {
				if (arr[fieldIndexer] == "@ECPGADD") add = 1;
				else commentstr=commentstr " " arr[fieldIndexer];
			} else {
				line=line " " arr[fieldIndexer]
			}
			continue
		}

		if (arr[fieldIndexer]=="}") {brace_indent--; continue}
		if (arr[fieldIndexer]=="{") {brace_indent++; continue}

		if (arr[fieldIndexer]=="/*" && brace_indent==0) {
			# start of a multiline comment
			comment=1;
			commentstr="/*"
			continue
		}

		if (arr[fieldIndexer]=="//" && brace_indent==0) {
			commentstr="//"
			for (cnt=fieldIndexer+1; cnt<=n; cnt++) {
				commentstr=commentstr arr[cnt];
			}
			add_to_buffer("rules", commentstr)
			next
		}

		if (brace_indent>0) { continue }

		if (arr[fieldIndexer]==";") {
			if (infield ) {
				dump_line(stmt_mode, fields, field_count, leading_spaces);
			}
			field_count=0;
			infield=0;
			add_to_buffer("rules",";\n\n")
			line=""
			continue;
		}


		if (arr[fieldIndexer]=="|") {
			if (infield ) {
				dump_line(stmt_mode, fields, field_count, leading_spaces);
			}
			field_count=0;
			infield=1;
			line= "| "
			continue;
		}

		if (replace_token[arr[fieldIndexer]]) {
			arr[fieldIndexer]=replace_token[arr[fieldIndexer]] 
		}

		if (line!="") {
			line=line " " arr[fieldIndexer]
		} else {
			line=arr[fieldIndexer]
		}
				
		# Are we looking at a declaration of a non-terminal ? 
		if ((arr[fieldIndexer] ~ "[A-Za-z0-9]+:") || arr[fieldIndexer+1]==":") {
			leading_spaces="        "
			non_term_id=arr[fieldIndexer]
			gsub(":","",non_term_id)

			# Do we have the : attached already ? 
			if (arr[fieldIndexer] ~ "[A-Za-z0-9]+:") {
				# We'll have already printed the ':'
			} else {
				# Consume the ':' which is next...
				line= line ":"
				fieldIndexer++;
			}

			if (non_term_id=="stmt") {
				# Special mode 
				stmt_mode=1;
			} else {
				stmt_mode=0;
			}

			if (types[non_term_id]=="") {
				types[non_term_id]="<str>";
			}

			tstr="%type " types[non_term_id] " " non_term_id
			add_to_buffer("types", tstr)

			line= line "/* RULE */";
			add_to_buffer("rules", line)
			line=""
			field_count=0;
			fields[0]="";
			fields[1]="";
			infield=1;
			continue;
		}


		if (arr[fieldIndexer]=="%prec") {
			prec = 1;
			continue;
		}

		if (!prec && !comment  && arr[fieldIndexer]!="/*EMPTY*/"  && length(arr[fieldIndexer]) && infield) {
			nfield=field_count+1;
			if (arr[fieldIndexer] != "Op" && (tokens[arr[fieldIndexer]]>0 || arr[fieldIndexer] ~ "'.+'")) {
				if (replace_string[arr[fieldIndexer]]) s = replace_string[arr[fieldIndexer]];
				else s=arr[fieldIndexer];
				gsub("_P","",s)
				gsub("'","",s)
				fields[field_count++]=tolower(s);
			} else {
				fields[field_count++]="$" nfield;
			}
		}
	}
}


END {
	dump("header")
	dump("tokens")
	dump("types")
	dump("ecpgtype")
	dump("orig_tokens")
	dump("rules")
	dump("trailer")
}
